import grp11.AVLPriorityHeapTrigramStorage;
import grp13.WeirdTrigramStorage;
import grp4.OMGOptimizedTrigramStorage;
import grp5.TrieTrigramStorage;
import grp8.Treap;
import grp9.MyTrigramStorage;
import i2.BinarySearchTrigramStorage;
import i2.OptimizedTrigramStorage;
import i2.SimpleTrigramStorage;

import java.io.IOException;

import trigrams.ImplementationFailException;
import trigrams.TrigramFileReader;
import trigrams.TrigramQueryReader;
import trigrams.TrigramStorage;

public class TrigramMain {

	private static void testStorage(TrigramStorage storage, String trigramFilename, String queryFilename, Integer runs, boolean buildOnly, Boolean verbose) {
		boolean verboseFlag = false;
		if (verbose != null) {
			verboseFlag = verbose;
		}
		try {
			TrigramFileReader tfr = new TrigramFileReader(trigramFilename, storage, verboseFlag);
		} catch (IOException e) {
			System.out.println("Error while reading the trigrams. Traceback:");
			e.printStackTrace();
			System.exit(-1);
		}
		
		// If we were asked to only build the storage, quit here.
		if (buildOnly) {
			return;
		}
		
		int totalRuns = 1;
		if (runs != null) {
			totalRuns = runs;
		}
		
		for (int run = 0; run < totalRuns; ++run) {
			try {
				TrigramQueryReader tqr = new TrigramQueryReader(queryFilename, storage, verboseFlag);
			} catch (IOException e) {
				System.out.println("Error while reading the queries. Traceback:");
				e.printStackTrace();
				System.exit(-1);
			} catch (ImplementationFailException e) {
				System.out.println(e.getMessage());
				System.exit(-1);
			}
		}
	}
	
	static void generateQueries(BinarySearchTrigramStorage storage, String trigramFilename, String queriesFilename, Boolean verbose) {
		boolean verboseFlag = false;
		if (verbose != null) {
			verboseFlag = verbose;
		}
		try {
			TrigramFileReader tfr = new TrigramFileReader(trigramFilename, storage, verboseFlag);
		} catch (IOException e) {
			System.out.println("Error while reading the trigrams. Traceback:");
			e.printStackTrace();
			System.exit(-1);
		}
		try {
			storage.generateQueries(queriesFilename);
		} catch (Exception e) {
			System.out.println("Error while generating queries. Traceback:");
			e.printStackTrace();
			System.exit(-1);
		}
	}
	
	public static void main(String[] args) {
		if (args.length < 2) {
			System.out.println("Wrong usage. Please supply the necessary parameters.");
			System.out.println();
			System.out.println("Usage: java TrigramMain [-vb] [-r<runs>] [-gen <queryFile>] [-i<0,1,2>] [-grp<4,5,8,9,11,13>] <TrigramFile> <QueryFile>");
			System.exit(-1);
		} else {
			Integer i2Implementation = null;
			Integer groupImplementation = null;
			Boolean verbose = null;
			Integer runs = null;
			String trigramFile = null;
			String queryFile = null;
			boolean buildOnly = false;
			Boolean generateQueries = null;
			String generateQueriesFile = null;
			
			for (int position = 0; position < args.length; ++position) {
				String arg = args[position];
				
				// If the option is a flag, treat it accordingly.
				if (arg.startsWith("-")) {
					if (arg.equals("-v")) {
						verbose = true;
					} else if (arg.startsWith("-i")) {
						i2Implementation = Integer.parseInt(arg.substring(2));
					} else if (arg.startsWith("-grp")) {
						groupImplementation = Integer.parseInt(arg.substring(4));
					} else if (arg.startsWith("-r")) {
						runs = Integer.parseInt(arg.substring(2));
					} else if (arg.startsWith("-b")) {
						buildOnly = true;
					} else if (arg.startsWith("-gen")) {
						generateQueries = true;
						++position;
						generateQueriesFile = args[position]; 
					}
				} else {
					// If it's a positional argument, we expect the trigram file first and then the query file.
					if (trigramFile == null) {
						trigramFile = arg;
					} else if (queryFile == null) {
						queryFile = arg;
					} else {
						System.out.println("Wrong usage. Too many positional arguments.");
						System.exit(-1);
					}
				}
			}
			
			if (groupImplementation != null && i2Implementation != null) {
				System.out.println("Wrong usage. Please specify either an i2 implementation *or* a group implementation.");
				System.exit(-1);
			}
			
			if (generateQueries != null && generateQueries == true) {
				generateQueries(new BinarySearchTrigramStorage(), trigramFile, generateQueriesFile, verbose);
			} else {
				if (groupImplementation != null) {
					switch (groupImplementation) {
					case 4: testStorage(new OMGOptimizedTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					case 5: testStorage(new TrieTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					case 8: testStorage(new Treap(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					case 9: testStorage(new MyTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					case 11: testStorage(new AVLPriorityHeapTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					case 13: testStorage(new WeirdTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					default: System.out.println("Please choose a valid implementation."); System.exit(-1);
					}
				} else if (i2Implementation != null) {
					switch (i2Implementation) {
					case 0: testStorage(new SimpleTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					case 1: testStorage(new BinarySearchTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					case 2: testStorage(new OptimizedTrigramStorage(), trigramFile, queryFile, runs, buildOnly, verbose); break;
					default: System.out.println("Please choose a valid implementation."); System.exit(-1);
					}
				} else {
					System.out.println("Please choose a valid implementation.");
					System.exit(-1);
				}
			}
		}
	}
}
