/* Gruppe 13
 *
 * Autoren: Leon Wittwer, Rene Schaefer, Johannes Seiffarth
 */
package grp13;

public class TrigramAndCountTree2LeonEdition2 {

	public static final int arraySize = 1 + 10 + 26;
	
	public class Node
	{
		Node parent, children[], minLeaf, maxLeaf;
		int iOffsetLeft = -1, iOffsetRight = -1;
		int iParentArrayIndex = -1;
		int iIndex;
		boolean bTrigram = false;
		
		public Node(int iIndex, Node parent)
		{
			this.iIndex = iIndex;
			this.parent = parent; 
			children = null;//new Node[arraySize];
		}
		
		public boolean hasChildren()
		{
			return iOffsetLeft > -1;
		}
		
		public void setChild(int iArrayPos, Node child)
		{
			if(children == null)
			{
				children = new Node[1];
				// set value
				children[0] = child;
				child.iParentArrayIndex = 0;
				// set the offsets				
				iOffsetLeft = iArrayPos;
				iOffsetRight = iArrayPos;
			}
			else if (!(iOffsetLeft <= iArrayPos && iArrayPos <= iOffsetRight)){
				// The array has to be extended
				
				Node[] newChildren;
				if(iArrayPos < iOffsetLeft)
				{	// Left enlarge
					newChildren = new Node[iOffsetRight - iArrayPos + 1];
					// copy old data
					for(int i = 0; i <= (iOffsetRight-iOffsetLeft); i++)
					{
						// if necessary
						if(children[i] != null)
						{
							newChildren[iOffsetLeft - iArrayPos + i] = children[i];
							children[i].iParentArrayIndex = iOffsetLeft - iArrayPos + i;
						}
					}
					// update offset
					iOffsetLeft = iArrayPos;
					// copy new array
					children = newChildren;
				}
				else if(iArrayPos > iOffsetRight)
				{	// right enlarge
					newChildren = new Node[iArrayPos - iOffsetLeft + 1];
					// copy old data
					for(int i = 0; i <= (iOffsetRight-iOffsetLeft); i++)
					{
						// if necessary
						if(children[i] != null)
						{
							newChildren[i] = children[i];
							children[i].iParentArrayIndex = i;
						}
					}
					// update offset
					iOffsetRight = iArrayPos;
					// copy new array
					children = newChildren;
				}
			}
			
			//set the value
			children[iArrayPos-iOffsetLeft] = child;
			child.iParentArrayIndex = iArrayPos-iOffsetLeft;
		}
		
		public Node getChild(int iArrayPos)
		{
			return children[iArrayPos - iOffsetLeft];
		}
		
		public boolean containsChild(int iArrayPos)
		{
			if(iOffsetLeft <= iArrayPos && iArrayPos <= iOffsetRight)
				return getChild(iArrayPos) != null;
			else return false;
		}
		
		public boolean containsIndex(int iArrayPos)
		{
			return (iOffsetLeft <= iArrayPos && iArrayPos <= iOffsetRight);
		}

	}
	
	public class SearchResult
	{
		public SearchResult(Node result)
		{
			res = result;
		}
		
		public SearchResult(Node temp, boolean searchFailure)
		{
			res = temp;
			this.searchFailure = searchFailure;
		}
		
		public Node res;
		public boolean searchFailure = false;
	}
	
	Node root = new Node(0, null);
	boolean bLastSearchApproximated = false;

	public boolean lastSearchApproximated()
	{
		return bLastSearchApproximated;
	}
	
	public void build()
	{
		build(root);
	}
	
	private void build(Node start)
	{
		
		if(start.iOffsetLeft == -1)
		{
			start.minLeaf = start;
			start.maxLeaf = start;
			return;
		}
		for(int i = 0; i <= start.iOffsetRight - start.iOffsetLeft; i++)
		{
			if(start.children[i] != null)
				build(start.children[i]);
		}
		
		if(start.bTrigram)
			start.minLeaf = start;
		else
			start.minLeaf = start.children[0].minLeaf;
			
		start.maxLeaf = start.children[start.iOffsetRight - start.iOffsetLeft].maxLeaf;
	}
	
	/**
	 * Searches a lower border for the trigram space. 
	 * @param tri to search for
	 * @param start is the start node
	 * @return the index of the
	 */
	public int leftSearch(String strPath)
	{
		bLastSearchApproximated = false;
		Node start = root;
		for(int j = 0; j < strPath.length(); j++)
		{
			// check whether there is a consisting node for the actual character
			int arrayPos = getArrayPos(strPath.charAt(j));
			// create the new node
			if(!start.containsChild(arrayPos))
			{ // the requested trigram is not in the tree. Give back a search failure compensation node that is not in the tree
	
				bLastSearchApproximated = true;
				// we are looking for a lower border for the trigram space
				
				if(start.children == null)
				{	// There are no children and the node has to represent a trigram
					return start.iIndex+1;
				}
				else
				{	// There are some children
					if(arrayPos > start.iOffsetRight)
					{	// there is a left child we'll use
						// our position is at the right side of the array
						
						// go to the left
//						return maxLeaf(start.children[start.iOffsetRight-start.iOffsetLeft]).iIndex+1;
						return start.children[start.iOffsetRight-start.iOffsetLeft].maxLeaf.iIndex+1;
					}
					else if(arrayPos < start.iOffsetLeft)
					{	// there is a right child we'll use
						// our position is left of the array
						
						// go to the right
//						return minLeaf(start.children[0]).iIndex;
						return start.children[0].minLeaf.iIndex;
					}
					else
					{
						// translation into node children charcater space
						int k = arrayPos - start.iOffsetLeft;
						// lookin for next left child
						for(; start.children[k] == null; k--);
						
//						return maxLeaf(start.children[k]).iIndex+1;
						return start.children[k].maxLeaf.iIndex+1;
					}
				}
			}
			// step downwards
			start = start.getChild(arrayPos);
		}
		
		
		// Check whether this start node is really a trigram and not just a crossing
		
		if(start.bTrigram)
			// the node represents a trigram and the index can be returned
			return start.iIndex;
		else
		{	// the node does not represent a trigram, so give back the outer most left child
//			return minLeaf(start.children[0]).iIndex;
			return start.children[0].minLeaf.iIndex;
		}
	}
	
	public int rightSearch(String strPath)
	{
		bLastSearchApproximated = false;
		Node start = root;
		for(int j = 0; j < strPath.length(); j++)
		{
			// check whether there is a consisting node for the actual character
			int arrayPos = getArrayPos(strPath.charAt(j));
			// create the new node
			if(!start.containsChild(arrayPos))
			{ // the requested trigram is not in the tree. Give back a search failure compensation node that is not in the tree
	
				bLastSearchApproximated = true;
				// we are looking for a lower border for the trigram space
				
				if(start.children == null)
				{	// There are no children
					return start.iIndex;
				}
				else
				{	// There are some children
					if(arrayPos > start.iOffsetRight)
					{	// there is a left child we'll use
						// our position is at the right side of the array
						
						// go to the left
//						return maxLeaf(start.children[start.iOffsetRight-start.iOffsetLeft]).iIndex;
						return start.children[start.iOffsetRight-start.iOffsetLeft].maxLeaf.iIndex;
					}
					else if(arrayPos < start.iOffsetLeft)
					{	// there is a right child we'll use
						// our position is left of the array
						
						// go to the right
//						return minLeaf(start.children[0]).iIndex-1;
						return start.children[0].minLeaf.iIndex-1;
					}
					else
					{
						// translation into node children charcater space
						int k = arrayPos - start.iOffsetLeft;
						// lookin for next left child
						for(; start.children[k] == null; k--);
						
//						return maxLeaf(start.children[k]).iIndex;
						return start.children[k].maxLeaf.iIndex;
					}
				}
			}
			// step downwards
			start = start.getChild(arrayPos);
		}
			
		// Check whether this start node is really a trigram and not just a crossing
		
		if(start.bTrigram)
			// the node represents a trigram and the index can be returned
			return start.iIndex;
		else
		{	// the node does not represent a trigram, so give back the outer most left child
//			return minLeaf(start.children[0]).iIndex-1;
			return start.children[0].minLeaf.iIndex-1;
		}
	}

	
	/**
	 * 
	 * @param a the left trigram
	 * @param b the right trigram
	 * @return the index of the trigram crossing or -1 in case of fail 
	 */
/*	public SearchResult searchCrossNode(Trigram a, Trigram b)
	{
		Node start = root;
		for(int i = 0; i < 3; i++)
		{
			String strPathA = null;
			String strPathB = null;
			switch(i)
			{
			case 0:
				strPathA = a.getW1();
				strPathB = b.getW1();
				break;
			case 1:
				strPathA = a.getW2();
				strPathB = b.getW2();
				break;
			case 2:
				strPathA = a.getW3();
				strPathB = b.getW3();
				break;
			}

			for(int j = 0; j < strPathA.length() && j < strPathB.length(); j++)
			{
				// check whether there is a consisting node for the actual character
				int arrayPosA = getArrayPos(strPathA.charAt(j));
				int arrayPosB = getArrayPos(strPathB.charAt(j));
				
				if(arrayPosA != arrayPosB)
				{	// There is a path crossing
					return new S;
				}
				
				// create the node existence
				if(!start.containsChild(arrayPosA))
				{ // the requested trigram is not in the tree, so give back the temp node with a search failure
					return new SearchResult(start, true);
				}
				// step downwards
				start = start.getChild(arrayPosA);
			}
			
			// They are going the same path for the word, but the word size is
			// different so this is definitely a cross node
			if(strPathA.length() != strPathB.length())
				return new SearchResult(start);
			
			if(i < 2)
			{
				// take the word ended path
				if(!start.containsChild(0))
				{ // the requested trigram is not in the tree so give back the temporary node with a search failure
					return new SearchResult(start, true);
				}
				start = start.children[0];
			}
		}
		
		return new SearchResult(start);
	}*/
	
	public void insert(String strPath, int iIndex)
	{
		Node start = root;
		
		for(int j = 0; j < strPath.length(); j++)
		{				
			// check whether there is a consisting node for the actual character
 			int arrayPos = getArrayPos(strPath.charAt(j));
			// create the new node
			if(!start.containsChild(arrayPos))
			{
				start.setChild(arrayPos, new Node(0, start));
			}
			// step downwards
			start = start.getChild(arrayPos);
		}
		
		// set the index
		if(start.iIndex != 0)
			System.out.println("Index already exists: " + iIndex);
		start.iIndex = iIndex;
		// set trigram representation
		start.bTrigram = true;
	}
		
	
	public static int getArrayPos(char c)
	{
		// order [epsilon,numbers,characters]
		if(c == ' ')
			return 0;
		else if('0' <= c && c <= '9')
		{
			// c is a number
			return 1 + c - '0';
		}
		else
		{
			// c is not a number but a character
			return 1 + 10 + c - 'a';
		}
	}

	public static Node minLeaf(Node start)
	{
		while(start.hasChildren() && !start.bTrigram)
		{
			start = start.children[0];
		}
		return start;
	}
	
	public static Node maxLeaf(Node start)
	{
		while(start.hasChildren())
		{
			// Last array position is always a child and the maximum in alphabet
			start = start.children[start.children.length-1];
		}
		return start;
	}
	
	private String pos(TrigramAndCountTree2LeonEdition2.Node node)
	{
		String res = "";
		
		while(node.parent != null)
		{
			TrigramAndCountTree2LeonEdition2.Node parent = node.parent;
			
			// search the child pos of start in parent
			res = arrayPosToChar(node.iParentArrayIndex + parent.iOffsetLeft) + res;
			
			node = node.parent;
		}
		
		return res;
	}
	private char arrayPosToChar(int i)
	{
		if(i == 0)
			return ' ';
		if(1 <= i && i <= 10)
			return (char)('0' + (i-1));
		else
			return (char)('a' + (i-11));
		
	}
	/**
	 * 
	 * @param start node of which we want to have the successor
	 * @return the successor of the node start or null if there is no one
	 */
	public static Node succ(Node start)
	{
		while(start.parent != null)
		{
			// Check wether there are siblings to the right
			for(int i = start.iParentArrayIndex; i <= (start.iOffsetRight - start.iOffsetLeft); i++)
			{
				if(start.parent.children[i] != null)
				{	// There is a sibling to the right
					// return the outer left leaf
					return minLeaf(start.parent.children[i]);
				}
			}
			start = start.parent;
		}
		
		// There is no successor. The node was at the right border
		return null;
	}
	
	/**
	 * 
	 * @param start node of which we want to have the predecessor
	 * @return the predecessor of the node start or null if there is no one
	 */
	public static Node predec(Node start)
	{
		while(start.parent != null)
		{
			for(int i = start.iParentArrayIndex; i >= 0; i--)
			{
				if(start.parent.children[i] != null)
				{
					return maxLeaf(start.parent.children[i]);
				}
			}
			start = start.parent;
		}
		
		// There is no predecessor. The node was the outer most left one. 
		return null;
	}

		
		
/*		int count = tri.getCount();
		
		// pass the tree with the first string
		Node firstString = searchStringToInsert(tri.getW1(), root, count);
		// check whether the "word end" child is non null
		if(firstString.children[0] == null)
		{
			firstString.children[0] = new Node(0, firstString);
		}
		// walk the tree with the second string
		Node secondString = searchStringToInsert(tri.getW2(), firstString.children[0], count);
		// make sure the "word end" child is non null
		if(secondString.children[0] == null)
		{
			secondString.children[0] = new Node(0, secondString);
		}
		// walk the tree with the third string
		Node destString = searchStringToInsert(tri.getW3(), secondString.children[0], count);
		
		// this is the node for the trigram
		// so copy the count
		destString.count = count;
	}
	
	private Node searchStringToInsert(String strPath, Node start, int countToInsert)
	{
		// loop over all characters in the string
		for(int i = 0; i < strPath.length(); i++)
		{
			// actualise the maximum count for the current node, because
			// countToInsert becomes its child
			start.checkMax(countToInsert);
			start.sum += countToInsert;
			
			// check whether there is a consisting node for the actual character
 			int arrayPos = strPath.charAt(i)-'a'+1;
			// create the new node
			if(start.children[arrayPos] == null)
			{
				start.children[arrayPos] = new Node(0, start);
			}
			// step downwards
			start = start.children[arrayPos];
		}
		
		return start;
	}*/
	
}
