/* Gruppe 13
 *
 * Autoren: Leon Wittwer, Rene Schaefer, Johannes Seiffarth
 */
package grp13;

import java.util.ArrayList;

import trigrams.*;

public class WeirdTrigramStorage implements TrigramStorage {

	public WeirdTrigramStorage() {
		trigrams = new ArrayList<TrigramAndCount>();
	}
	
	@Override
	public void insert(TrigramAndCount trigram) {
		//just get all Trigrams
		trigrams.add(trigram);
	}

	@Override
	public void build() {
		//working with all trigrams
		//put them in two arrays, one for the trigram and one for the count
		nroftrigrams = trigrams.size();
		trigrams1 = new String[nroftrigrams];
		counts1 = new int[nroftrigrams];
		for ( int i=0; i<nroftrigrams; i++ )
		{
			TrigramAndCount tr = trigrams.get(i);
			trigrams1[i] = new String( tr.getW1() + " " + tr.getW2() + " " + tr.getW3() );
			counts1[i] = tr.getCount();
		}
		//sort the arrays of all trigrams
		heapSort(trigrams1, counts1);
		
		//create array for sums, treap for max, hashtable, and trie
		countsum = new long[nroftrigrams];
		treap4max = new Treap();
		trigrams3 = new MyHashMap(nroftrigrams);
		trie = new TrigramAndCountTree2LeonEdition2();
		//fill them
		for ( int i=0; i<nroftrigrams; i++ )
		{
			countsum[i] = counts1[i];
			treap4max.insert( i, counts1[i] );
			trigrams3.put( trigrams1[i], i );
			trie.insert( trigrams1[i], i );
		}
		trie.build();
		
		//precalculate all sums
		for ( int i=1; i<nroftrigrams; i++ ) {
			countsum[i] = countsum[i-1] + countsum[i];
		}
	}
	
	@Override
	public QueryResult query(Trigram trigram1, Trigram trigram2) {
		long max = 0;
		long sum = 0;
		
		//get query trigrams as strings
		String tr1 = new String( trigram1.getW1() + " " + trigram1.getW2() + " " + trigram1.getW3() );
		String tr2 = new String( trigram2.getW1() + " " + trigram2.getW2() + " " + trigram2.getW3() );
		
		//get the indices for trigram1 and 2 with hashing (or searching in trie)
		int index1 = -1;
		int index2 = -1;
		int start = 0, end = nroftrigrams-1;
		//bools to get the case of the empty answer set
		boolean bool1 = false, bool2 = false;
		
		Integer i1 = trigrams3.get( tr1 );
		if (i1 != null) {
			index1 = (int)i1;
		}
		else {
			index1 = trie.leftSearch(tr1);
			bool1 = trie.lastSearchApproximated();
		}
		Integer i2 = trigrams3.get( tr2 );
		if (i2 != null) {
			index2 = (int)i2;
		}
		else {
			index2 = trie.rightSearch(tr2);
			bool2 = trie.lastSearchApproximated();
		}
		
		//cases empty set or just one element
		if ( bool1 && bool2 )
		{
			if ( index1 > index2 )
				return new QueryResult(0, 0);
			else if ( index1 == index2 )
				return new QueryResult( counts1[index1], counts1[index1] );
		}
		
		//calculate sum
		if ( index1 == 0 ) {
			sum = countsum[index2];
		}
		else {
			sum = countsum[index2] - countsum[index1-1];
		}
		
		//get max
		TreapNode cur = treap4max.root;
		while (true)
		{
			if ( index1 == cur.key || index2 == cur.key || (index1 < cur.key && index2 > cur.key) ) {
				max = cur.prio;
				break;
			}
			else if ( index1 < cur.key ) {
				cur = cur.left;
			}
			else {
				cur = cur.right;
			}
		}
		
		return new QueryResult(max, sum);
	}
	
	//heapsort implementation for sorting the two arrays
	void buildHeap( String E[], int counts[] ) {
		for (int i = E.length / 2 - 1; i >= 0; i--) {
			heapify(E, counts, E.length, i);
		}
	}
	void heapify( String E[], int counts[], int n, int pos ) {
		int next = 2 * pos + 1;
		while (next < n) {
			if (next + 1 < n && E[next + 1].compareTo(E[next]) > 0 ) {
				next = next + 1;
			}
			if (E[pos].compareTo(E[next]) >= 0) {
				break;
			}
			swap( E, counts, pos, next );
			pos = next;
			next = 2 * pos + 1;
		}
	}
	void heapSort( String E[], int counts[] ) {
		buildHeap( E, counts );
		for (int i = E.length - 1; i > 0; i--) {
			swap( E, counts, 0, i );
			heapify(E, counts, i, 0);
		}
	}
	void swap( String E[], int counts[], int a, int b ) {
		String ts = E[a];
		E[a] = E[b];
		E[b] = ts;
		int tc = counts[a];
		counts[a] = counts[b];
		counts[b] = tc;
	}
	
	protected TrigramAndCountTree2LeonEdition2 trie;
	protected ArrayList<TrigramAndCount> trigrams;
	protected String[] trigrams1;//strings of trigrams(ordered)
	protected MyHashMap trigrams3;//hashmap for indices of trigrams for faster finding indices
	protected int[] counts1;//counts of trigrams(ordered)
	protected long[] countsum;//pre-calculated sums of trigrams
	protected int nroftrigrams;
	protected Treap treap4max;//treap to find the max count between the trigrams
	
}
