/* Gruppe 4
 *
 * Autoren: anonym
 */
package grp4;

import java.util.ArrayList;

import trigrams.*;

/*
Idea: remember maxes and sums n-arily
    so it takes log n time instead of n time to calculate
    so we need a binary sorted structure (loglib, now maxlib and sumlib)
    how do we store the 'meta-information'?
        we don't, our structure never changes so it's ok to do statically
    we n-arily find start and end, then n-arily compute sum and max

"Special optimization":
    the main loop(s) take < 2 ms per run, the binary search via
        compareTo takes > 20 ms per run
    so we store a rank (rk) of type long for every trigram with the property
        rk(x) < rk(y)  =>  x.compareTo(y) < 0
    this is easily possible as compareTo only compares x and y's unicode values
        (not using anything tricky like locales or a = A < b)
    so we only have to call the 'slow' compareTo function if rk(x) = rk(y)
    now that part only takes ~ 12 ms per run
*/

public class OMGOptimizedTrigramStorage implements TrigramStorage {

    public OMGOptimizedTrigramStorage() {
        trigrams = new ArrayList<TrigramAndCount>(1000000);
    }

    @Override
    public void insert(TrigramAndCount trigram) {
        trigrams.add(trigram);
    }


    int partition(int left, int right) {
        int ppos = right;
        Trigram pivot = trigrams.get(ppos);
        while (true) {
            while (left < right)
                if (rklib[left] < rklib[ppos])  left++;
                else if (rklib[left] > rklib[ppos])  break;
                else if ((trigrams.get(left)).compareTo(pivot) < 0)  left++;
                else break;
            while (left < right)
                if (rklib[right] > rklib[ppos])  right--;
                else if (rklib[right] < rklib[ppos])  break;
                else if ((trigrams.get(right)).compareTo(pivot) >= 0)  right--;
                else break;
            if (left >= right)  break;
            TrigramAndCount tmp = trigrams.get(left); trigrams.set(left, trigrams.get(right)); trigrams.set(right, tmp);
            long tmprk = rklib[left];  rklib[left] = rklib[right];  rklib[right] = tmprk;
        }
        TrigramAndCount tmp = trigrams.get(left); trigrams.set(left, trigrams.get(ppos)); trigrams.set(ppos, tmp);
        long tmprk = rklib[left];  rklib[left] = rklib[ppos];  rklib[ppos] = tmprk;
        return left;
    }

    void quickSort(int left, int right) {
        if (left < right) {
            int i = partition(left, right);
            quickSort(left, i - 1);
            quickSort(i + 1, right);
        }
    }



    long rk(Trigram tri) {
        String w = tri.getW1() + "/" + tri.getW2() + "/" + tri.getW3();
        long rk = 0;
        switch (w.length()) {
            default: rk += ((long)w.charAt(9) - 47L);
            case 9:  rk += ((long)w.charAt(8) - 47L) * 76L;
            case 8:  rk += ((long)w.charAt(7) - 47L) * 5776L;
            case 7:  rk += ((long)w.charAt(6) - 47L) * 438976L;
            case 6:  rk += ((long)w.charAt(5) - 47L) * 33362176L;
            case 5:  rk += ((long)w.charAt(4) - 47L) * 2535525376L;
            case 4:  rk += ((long)w.charAt(3) - 47L) * 192699928576L;
            case 3:  rk += ((long)w.charAt(2) - 47L) * 14645194571776L;
            case 2:  rk += ((long)w.charAt(1) - 47L) * 1113034787454976L;
            case 1:  rk += ((long)w.charAt(0) - 47L) * 84590643846578176L;
        }
        // 'switch' using hardcoded longs ^= 16ms, 'for' using pow ^= 23 ms !
        // ^remember we have to calculate rk only /once per query/

        return rk;
    }

    void build_loglib() {
        int m = n / arity;
        int currun = arity;

        int logn = 0;  // log(n) + 1 actually
        for (int c = 1; c <= n; c *= arity) { logn++; }


        rklib = new long[n];

        for (int i = 0; i < n; ++i)
            rklib[i] = rk(trigrams.get(i));


        quickSort(0, n-1);


        sumlib = new long[logn][];
        maxlib = new long[logn][];
        rklib = new long[n];

        sumlib[0] = new long[n];
        maxlib[0] = new long[n];

        for (int i = 0; i < n; ++i) {
            TrigramAndCount x = trigrams.get(i);
            sumlib[0][i] = x.getCount();
            maxlib[0][i] = x.getCount();
            rklib[i] = rk(x);
        }


        for (int j = 1; ; ++j) {

            if (currun > n) { break; }

            m = n / currun;
            sumlib[j] = new long[m];
            maxlib[j] = new long[m];

            for (int i = 0; i < m; ++i) {

                long max = 0;
                for (int k = 0; k < arity; k++)
                    if (maxlib[j-1][arity*i+k] > max)
                        max = maxlib[j-1][arity*i+k];
                long sum = 0;
                for (int k = 0; k < arity; k++)
                    sum += sumlib[j-1][arity*i+k];
                maxlib[j][i] = max;
                sumlib[j][i] = sum;
            }

            // ---
            maxrun = currun;
            maxlvl = j;

            currun *= arity;
        }
    }


    @Override
    public void build() {
        n = trigrams.size();
        build_loglib();
    }


    int binsearch(Trigram tri, int lo, boolean is_high) {
        long rk = rk(tri);
        int hi = n-1;
        while (lo <= hi) {
            int mid = lo + (hi - lo) / 2;
            if      (rk < rklib[mid]) hi = mid - 1;
            else if (rk > rklib[mid]) lo = mid + 1;
            else {
                int javacmp = tri.compareTo(trigrams.get(mid));  // sloooow!
                if (javacmp < 0) hi = mid - 1;
                else if (javacmp > 0) lo = mid + 1;
                else return mid;
            }
        }
        if (is_high)
            return hi;
        return lo;
    }

    @Override
    public QueryResult query(Trigram trigram1, Trigram trigram2) {

        int start = binsearch(trigram1, 0, false);
        int end = 1 + binsearch(trigram2, start, true);

        int diff = end - start;

        long sum = 0;
        long max = 0;

        if (diff < arity) {
            for (int i = start; i < end; ++i) {
                max = Math.max(max, maxlib[0][i]);
                sum += sumlib[0][i];
            }
            return new QueryResult(max, sum);
        }

        int currun = maxrun;
        int curlvl = maxlvl;
        while (currun > diff) { currun /= arity; curlvl--; }

        int divide = (end / currun) * currun;
        int divide_ = divide;

        int currun_ = currun;
        int curlvl_ = curlvl;

        currun /= arity;
        curlvl--;


        // left side
        diff = divide_ - start;
        while (currun_ > diff) { currun_ /= arity; curlvl_--; }

        while (currun_ >= arity) {
            int test = diff / currun_;
            divide_ = divide_ - test * currun_;
            for (int i = 0; i < test; ++i) {
                sum += sumlib[curlvl_][divide_ / currun_ + i];
                max = Math.max(max, maxlib[curlvl_][divide_ / currun_ + i]);
            }
            diff = divide_ - start;
            while (currun_ > diff) { currun_ /= arity; curlvl_--; }
        }

        for (; start < divide_; ++start) {
            max = Math.max(max, maxlib[0][start]);
            sum += sumlib[0][start];
        }


        // right side
        diff = end - divide;
        while (currun > diff) { currun /= arity; curlvl--; }

        while (currun >= arity) {
            int test = diff / currun;
            for (int i = 0; i < test; ++i) {
                sum += sumlib[curlvl][divide / currun + i];
                max = Math.max(max, maxlib[curlvl][divide / currun + i]);
            }
            divide = divide + test * currun;
            diff = end - divide;
            while (currun > diff) { currun /= arity; curlvl--; }
        }

        for (; divide < end; ++divide) {
            max = Math.max(max, maxlib[0][divide]);
            sum += sumlib[0][divide];
        }

        return new QueryResult(max, sum);
    }



    protected ArrayList<TrigramAndCount> trigrams;
    int n;

    final int arity = 2;
    long[][] maxlib;
    long[][] sumlib;
    long[] rklib;
    int maxrun;
    int maxlvl;
}
