/* Gruppe 5
 *
 * Autoren: anonym
 */
package grp5;

import java.util.Arrays;

import trigrams.TrigramAndCount;

public class ArrayInnerTrieNode extends InnerTrieNode {
	public TrieNode[][] nodes;
	public char[] table;

	public ArrayInnerTrieNode(String prefix) {
		super(prefix);
		this.nodes = new TrieNode[1][1]; //Zweidimensionales Array mit fast konstantem Zugriff
		this.table = new char[]{ 0 };
	}
	
	public void add(String key, TrieNode value){
		if("".equals(key)){
			if(nodes[0][0] != null){
				((TrieLeaf) nodes[0][0]).count += ((TrieLeaf) value).count;
			} else {
				nodes[0][0] = value;
			}
		} else {
			//Wir vergleichen nur den ersten character.
			char start = key.charAt(0);
			for(int i = 1; i < table.length; i++){
				//cur ist der berechnete Index fuer das aktuelle Teilarray
				int cur = start - table[i];
				
				if(cur < nodes[i].length){
					if(cur < 0){ //Index ist nicht vorhanden, also fuege passend ein
						if(cur == -1){
							//System.out.println("Fall 1: " + value);
							//vorne anhaengen
							nodes[i] = addFirst(nodes[i], value);
							table[i] = start;
						} else if(start - table[i-1] == nodes[i-1].length){
							//System.out.println("Fall 2a: " + value);
							//hinten anhaengen
							nodes[i] = addLast(nodes[i], value);
						} else {
							//System.out.println("Fall 3: " + value);
							//neues Array anlegen
							nodes = addAt(nodes, new TrieNode[]{value}, i);
							table = addAt(table, start, i);
						}
					} else {
						if(this.nodes[i][cur].prefix.equals(key)) return;
						//System.out.println("Fall 4: " + value);
						//Index ist vorhanden - splitte Knoten
						TrieNode toSplit = nodes[i][cur];
						//Berechne gemeinsame laenge der beiden Schluessel
						int l = getSameLength(key, toSplit.prefix);
						
						//Aendere die Praefixe der Knoten
						toSplit.prefix = toSplit.prefix.substring(l);
						value.prefix = key.substring(l);
						
						//Fuege neuen Knoten ein
						nodes[i][cur] = new ArrayInnerTrieNode(key.substring(0, l));
						
						//Fuege die beiden kollidierenden Knoten im neuen Knoten ein
						((InnerTrieNode) nodes[i][cur]).add(toSplit);
						((InnerTrieNode) nodes[i][cur]).add(value);
					}
					fix();
					return;
				} else if(cur == nodes[i].length){
					//System.out.println("Fall 2b: " + value);
					//hinten anhaengen
					nodes[i] = addLast(nodes[i], value);
					fix();
					return;
				}
			}
			//System.out.println("Fall 5: " + value);
			//neues array am Ende des bisherigen Bereichs anlegen
			nodes = addAt(nodes, new TrieNode[]{value}, nodes.length);
			table = addAt(table, value.prefix.charAt(0), table.length);
			fix();
		}
	}
	
	@Override
	public void add(TrieNode node) {
		add(node.prefix, node);
	}

	@Override
	public void add(TrigramAndCount t, String prefix) {
		add(prefix, new TrieLeaf(prefix, t.getCount(), 0));
	}

	@Override
	public void add(TrigramAndCount t) {
		if(nodes[0][0] != null){
			((TrieLeaf) nodes[0][0]).count += t.getCount();
		} else {
			nodes[0][0] = new TrieLeaf("", t.getCount(), 0);
		}
	}
	
	public TrieNode next(char[] search, int offset){
		if(offset == search.length){
			if(nodes[0][0] != null) return nodes[0][0];
			return (nodes.length > 1) ? nodes[1][0] : null;
		}
		char start = search[offset];
		for(int i = 1; i < table.length; i++){
			int index = start - table[i];
			if(0 > index) return null;
			if(index < nodes[i].length){
				return (startsWith(search, offset, nodes[i][index].prefix)) ? nodes[i][index] : null;
			}
		}
		return null;
	}

	@Override
	public TrieNode next(String key) {
		if("".equals(key)){
			if(nodes[0][0] != null) return nodes[0][0];
			return (nodes.length > 1) ? nodes[1][0] : null;
		}
		char start = key.charAt(0);
		for(int i = 1; i < table.length; i++){
			int index = start - table[i];
			if(0 > index) return null;
			if(index < nodes[i].length){
				return (key.startsWith(nodes[i][index].prefix)) ? nodes[i][index] : null;
			}
		}
		return null;
	}

	@Override
	public TrieNode nextCeil(String key) {
		if("".equals(key)){ //leerer String ist kleiner als alles andere
			if(nodes[0][0] == null) return (nodes.length > 1) ? nodes[1][0] : null;
			return nodes[0][0];
		}
		char start = key.charAt(0);
		//Suche nach Eintrag mit gleichem Anfangsbuchstaben
		for(int i = 1; i < table.length; i++){
			int index = start - table[i];
			if(0 > index) return nodes[i][0];
			if(index < nodes[i].length){
				if(key.startsWith(nodes[i][index].prefix) || nodes[i][index].prefix.compareTo(key) >= 0){
					return nodes[i][index];
				} else if(index + 1 < nodes[i].length){
					return nodes[i][index + 1];
				}
			}
		}
		return null;
	}

	@Override
	public TrieNode nextFloor(String key) {
		if("".equals(key)){
			return nodes[0][0];
		}
		char start = key.charAt(0);
		for(int i = table.length - 1; i > 0; i--){
			int index = start - table[i];
			if(index >= 0 && index < nodes[i].length){
				if(key.startsWith(nodes[i][index].prefix) || nodes[i][index].prefix.compareTo(key) <= 0){
					return nodes[i][index];
				} else {
					return (index > 0) ? nodes[i][index - 1] : nodes[i-1][nodes[i-1].length - 1];
				}
			} else if(index >= 0){
				return (index > 0) ? nodes[i][nodes[i].length - 1] : nodes[i-1][nodes[i-1].length - 1];
			}
		}
		return nodes[0][0];
	}

	@Override
	public TrieNode nextGreater(String key) {
		if(key == null){ //leerer String ist kleiner als alles andere (Ausser null)
			if(nodes[0][0] == null) return (nodes.length > 1) ? nodes[1][0] : null;
			return nodes[0][0];
		}
		if("".equals(key)){
			return (nodes.length > 1) ? nodes[1][0] : null;
		}
		char start = key.charAt(0);
		//Suche nach Eintrag mit gleichem Anfangsbuchstaben
		for(int i = 1; i < table.length; i++){
			int index = start - table[i];
			if(0 > index) return nodes[i][0];
			if(index < nodes[i].length){
				if(nodes[i][index].prefix.compareTo(key) > 0){
					return nodes[i][index];
				} else if(index + 1 < nodes[i].length){
					return nodes[i][index + 1];
				}
			}
		}
		return null;
	}

	@Override
	public TrieNode getFirst() {
		if(nodes[0][0] != null) return nodes[0][0];
		if(nodes.length > 1){
			return nodes[1][0];
		}
		return null;
	}
	
	@Override
	public TrieNode getLast(){
		int i = nodes.length - 1;
		return nodes[i][nodes[i].length - 1];
	}
	
	@Override
	public String toString(){
		return "(\"" + this.prefix + "\" -> " + Arrays.deepToString(nodes) + ")";
	}
	
	//Hilfsmethoden
	
	//Fuege zusammengehoerende Teilarrays zusammen und aktualisiere die lookup table
	private void fix(){
		for(int i = nodes.length - 1; i > 0; i--){
			if(table[i] - table[i-1] == nodes[i-1].length){
				nodes[i-1] = concat(nodes[i-1], nodes[i]);
				nodes = remove(nodes, i);
				table = remove(table, i);
				return;
			}
		}
	}
	
	//entfernt index aus dem array und shiftet rechts davon nach links.
	private static char[] remove(char[] a, int index){
		char[] res = new char[a.length - 1];
		System.arraycopy(a, 0, res, 0, index);
		if(index + 1 < a.length) System.arraycopy(a, index+1, res, index, a.length - index - 1);
		return res;
	}
	
	private static TrieNode[][] remove(TrieNode[][] a, int index){
		TrieNode[][] res = new TrieNode[a.length - 1][];
		System.arraycopy(a, 0, res, 0, index);
		if(index + 1 < a.length) System.arraycopy(a, index+1, res, index, a.length - index - 1);
		return res;
	}
	
	//Fuegt value (v) zu array (a) hinzu und shiftet bestehende Eintraege nach rechts. 
	private static char[] addAt(char[] a, char v, int index){
		char[] res = new char[a.length + 1];
		System.arraycopy(a, 0, res, 0, index);
		res[index] = v;
		System.arraycopy(a, index, res, index+1, a.length - index);
		return res;
	}
	
	private static TrieNode[][] addAt(TrieNode[][] a, TrieNode[] v, int index){
		TrieNode[][] res = new TrieNode[a.length + 1][];
		System.arraycopy(a, 0, res, 0, index);
		res[index] = v;
		System.arraycopy(a, index, res, index+1, a.length - index);
		return res;
	}

	//Konkateniere zwei arrays
	private static TrieNode[] concat(TrieNode[] a, TrieNode[] b){
		TrieNode[] res = new TrieNode[a.length + b.length];
		System.arraycopy(a, 0, res, 0, a.length);
		System.arraycopy(b, 0, res, a.length, b.length);
		return res;
	}
			
	//Fuege vorne an ein Array an
	private static TrieNode[] addFirst(TrieNode[] a, TrieNode t){
		TrieNode[] res = new TrieNode[a.length + 1];
		res[0] = t;
		System.arraycopy(a, 0, res, 1, a.length);
		return res;
	}
			
	//Fuege hinten an ein Array an
	private static TrieNode[] addLast(TrieNode[] a, TrieNode t){
		TrieNode[] res = new TrieNode[a.length + 1];
		System.arraycopy(a, 0, res, 0, a.length);
		res[a.length] = t;
		return res;
	}
	
	//True gdw. s2 ist Praefix von s1
	private static boolean startsWith(char[] s1, int offset, String s2){
		if(s1.length - offset < s2.length()) return false;
		int i = 0;
		int max = s2.length();
		while(i < max){
			if(s2.charAt(i) != s1[offset + i]) return false;
			i++;
		}
		return true;
	}
		
	//Berechne die gemeinsame Laenge der Strings
	private static int getSameLength(String s1, String s2){
		int i = 0;
		int max = Math.min(s1.length(), s2.length());
		while(i < max && s1.charAt(i) == s2.charAt(i)) i++;
		return i;
	}
}
