/* Gruppe 5
 *
 * Autoren: anonym
 */
package grp5;

import java.util.Iterator;

import trigrams.Trigram;
import trigrams.TrigramAndCount;

public class Trie implements Iterable<TrieLeaf>{
	ArrayInnerTrieNode root;
	public boolean ordered;
	
	public Trie(){
		this.root = new ArrayInnerTrieNode("");
		this.ordered = false;
	}
	
	public void insert(TrigramAndCount t){
		char[] search = TrigramWrapper.getChars(t);
		int index = 0; //Indiziert den Teilstring, der noch betrachtet wird
		TrieNode cur = this.root;
		TrieNode parent = null;
		while(cur instanceof InnerTrieNode){
			int i = 0;
			int max = (cur.prefix.length() < search.length - index) ? cur.prefix.length() : search.length - index;
			while(i < max && cur.prefix.charAt(i) == search[index + i]) i++;
			index += i;
			
			parent = cur;
			if(index >= search.length){
				//Es ex. ein String, dessen Praefix search ist -> fuege t mit leerem String ein
				((ArrayInnerTrieNode) cur).add(t);
				return;
			}
			cur = ((ArrayInnerTrieNode) cur).next(search, index);
		}
		String s = new String(search, index, search.length - index);
		((ArrayInnerTrieNode) parent).add(s, new TrieLeaf(s, t.getCount(), 0));
	}
	
	public TrieNode getNodeAt(Trigram t){
		return getNodeAt(TrigramWrapper.getString(t));
	}
	
	public TrieNode getNodeAt(String s){
		TrieNode cur = this.root;
		while(cur instanceof InnerTrieNode){
			cur = ((InnerTrieNode) cur).next(s);
			
			s = s.substring(cur.prefix.length());
			if("".equals(s)) return cur;
		}
		return cur;
	}
	
	public TrieNode getCeilingNode(Trigram t){
		return getCeilingNode(TrigramWrapper.getString(t));
	}
	
	public TrieNode getCeilingNode(String s){
		TrieNode cur = this.root;
		TrieNode parent = null;
		while(cur instanceof InnerTrieNode){
			//System.out.println(cur.prefix + " -> " + s);
			if(s.startsWith(cur.prefix)){
				s = s.substring(cur.prefix.length());
			} else {
				break;
			}
			parent = cur;
			cur = ((InnerTrieNode) cur).nextCeil(s);
		}
		if(cur == null){
			//System.out.println("p:" + parent);
			while(parent instanceof InnerTrieNode) parent = ((InnerTrieNode) parent).getLast();
			return (parent != null) ? ((TrieLeaf) parent).next : null;
		} else if(cur instanceof TrieLeaf){
			//System.out.println(parent);
			return (s.compareTo(cur.prefix) > 0) ? ((TrieLeaf) cur).next : cur;
		} else {
			//System.out.println("c:" + cur);
			while(cur instanceof InnerTrieNode) cur = ((InnerTrieNode) cur).getFirst();
			return cur;
		}
	}
	
	public TrieNode getFloorNode(Trigram t){
		return getFloorNode(TrigramWrapper.getString(t));
	}
	
	public TrieNode getFloorNode(String s){
		TrieNode cur = this.root;
		TrieNode parent = null;
		while(cur instanceof InnerTrieNode){
			//System.out.println(cur.prefix + " -> " + s);
			if(s.startsWith(cur.prefix)){
				s = s.substring(cur.prefix.length());
			} else {
				break;
			}
			parent = cur;
			cur = ((InnerTrieNode) cur).nextFloor(s);
		}
		if(cur == null){
			//System.out.println("p:" + parent);
			while(parent instanceof InnerTrieNode) parent = ((InnerTrieNode) parent).getFirst();
			return (parent != null) ? ((TrieLeaf) parent).prev : null;
		} else if(cur instanceof TrieLeaf){
			return (s.compareTo(cur.prefix) < 0) ? ((TrieLeaf) cur).prev : cur;
		} else {
			//System.out.println("c:" + cur);
			while(cur instanceof InnerTrieNode) cur = ((InnerTrieNode) cur).getLast();
			return cur;
		}
	}
	
	public TrieNode getRoot(){
		return this.root;
	}
	
	public TrieLeaf getMax(){
		TrieNode max = this.root;
		while(max instanceof InnerTrieNode){
			max = ((InnerTrieNode) max).getFirst();
		}
		return (TrieLeaf) max;
	}
	
	public TrieLeaf getMin(){
		TrieNode min = this.root;
		while(min instanceof InnerTrieNode){
			min = ((InnerTrieNode) min).getFirst();
		}
		return (TrieLeaf) min;
	}
	
	public Iterator<TrieLeaf> iterator(){
		return (ordered) ? new OrderedTrieIterator(this) : new TrieCursor(this);
	}
	
	public String toString(){
		return this.root.toString();
	}
	
	class OrderedTrieIterator implements Iterator<TrieLeaf>{
		TrieLeaf next;
		
		public OrderedTrieIterator(Trie t){
			this.next = t.getMin();
		}
		
		public boolean hasNext(){
			return (next != null);
		}
		
		public TrieLeaf next(){
			TrieLeaf cur = next;
			next = next.next;
			return cur;
		}

		public void remove() {
			throw new UnsupportedOperationException();
		}
	}
}
