/* Gruppe 5
 *
 * Autoren: anonym
 */
package grp5;

import java.util.Iterator;

public class TrieCursor implements Iterator<TrieLeaf>{
	//Der Stack, der den aktuellen Pfad speichert
	Stack<InnerTrieNode> stack;
	//Die Spitze des Pfades
	InnerTrieNode current;
	//Der String, der zum Blatt zuletzt ausgegebenen Blatt gehoert
	String string;
	//Der String, der zum naechsten Blatt gehoert
	StringBuilder nextString;
	//Das naechste Blatt, das ausgegeben wird.
	TrieLeaf next;
	
	public TrieCursor(Trie t){
		if(t != null){
			//Wir steigen ab zum kleinsten enthaltenen String.
			stack = new Stack<InnerTrieNode>();
			current = ((InnerTrieNode) t.root);
			TrieNode n = current.getFirst();
			string = "";
			nextString = new StringBuilder();
			
			while(n instanceof InnerTrieNode){
				nextString.append(n.prefix);
				stack.push(current);
				current = (InnerTrieNode) n;
				n = current.getFirst();
			}
			
			nextString.append(n.prefix);
			next = (TrieLeaf) n;
		} else {
			next = null;
			string = "";
		}
	}
	
	public String getString(){
		return this.string;
	}
	
	@Override
	public boolean hasNext(){
		return (next != null);
	}
	
	@Override
	public TrieLeaf next(){
		//Suche naechst groesseren String im Trie
		TrieLeaf res = next;
		string = nextString.toString();
		TrieNode n = current.nextGreater(next.prefix);
		
		//Kuerze den String um das Praefix von dem jetzt zurueckgegebenen Blatt
		nextString.delete(nextString.length() - next.prefix.length(), nextString.length());
		
		//n ist Blatt -> wir haben gefunden, was wir suchten
		if(n instanceof TrieLeaf){
			nextString.append(n.prefix);
			next = (TrieLeaf) n;
			return res;
		}
		
		//Es ex. kein Element mit lexikografisch groesserem Praefix an current
		while(n == null){
			if(stack.isEmpty()){
				//Stack ist leer, wir koennen nicht zurueck
				current = null;
				next = null;
				nextString.delete(0, nextString.length());
				return res;
			}
			//Kuerze den String
			String p = current.prefix;
			nextString.delete(nextString.length() - p.length(), nextString.length());
			
			//Nehme oberstes Element vom Stack und finde naechstgroesseren String
			current = stack.pop();
			
			n = current.nextGreater(p);
		}
		
		//n ist innerer Knoten -> lege current auf den Stack und finde kleinstes Blatt
		while(n instanceof InnerTrieNode) {
			nextString.append(n.prefix);
			stack.push(current);
			current = (InnerTrieNode) n;
			n = current.getFirst();
		}
		
		//Jeder Innere Knoten hat auch min. ein Blatt -> n ist hier ein Blatt
		next = (TrieLeaf) n;
		nextString.append(next.prefix);
		return res;
	}
	
	class Stack<V> {
		StackEntry<V> top;
		
		public Stack(){
			this.top = null;
		}
		
		public Stack(V top){
			this.top = new StackEntry<V>(top, null);
		}
		
		public boolean isEmpty(){
			return (top == null);
		}
		
		public void push(V v){
			this.top = new StackEntry<V>(v, this.top);
		}
		
		public V pop(){
			if(top != null){
				V v = top.v;
				top = top.next;
				return v;
			}
			return null;
		}
		
		public V peek(){
			if(top != null){
				return top.v;
			}
			return null;
		}
		
		@SuppressWarnings("hiding")
		class StackEntry<V> {
			final V v;
			final StackEntry<V> next;
			
			public StackEntry(V v, StackEntry<V> next){
				this.v = v;
				this.next = next;
			}
		}
	}

        public void remove() {
                throw new UnsupportedOperationException();
        }
}
