/* Gruppe 5
 *
 * Autoren: anonym
 */
package grp5;

import trigrams.*;

public class TrieTrigramStorage implements TrigramStorage {
	private Trie trie;
	private HashTable leaves;
	
	private int tricount;
	
	//maximale Anzahl der Sektoren
	private final int numOfSectors = 5000;
	//speichert die hoechsten Werte
	private int maxima[] = new int[numOfSectors-1];
	//gibt das Maximum zwischen 2 Sektoren wieder
	int sectorMax[][];
	
	public TrieTrigramStorage(){
		trie = new Trie();
	}
	
	public void insert(TrigramAndCount trigram){
		trie.insert(trigram);
		if(trigram.getCount() > maxima[0])
			sectorInsert(maxima, trigram.getCount());
		tricount++;
	}
	
	public void build(){
		leaves = new HashTable(tricount/7);
		System.out.println("Building hashtable, " + tricount/20 + " Entries");
		
		TrieLeaf prev = null;
		long sum = 0;
		int sector = 0;
		
		int sectorVal[] = new int[numOfSectors];
		//Durchlaufen von Vorne nach Hinten
		TrieCursor c = new TrieCursor(trie);
		TrieLeaf cur;

		while(c.hasNext()){
			cur = c.next();
			leaves.insert(cur, c.getString());
			
			sum += cur.count;
			cur.sum = sum;
			cur.prev = prev;
			
			if(cur.count > maxima[0]){
				//der naechste Sektor wird erstellt
				sector++;
				sectorVal[sector] = cur.count;
				cur.secOffset = 1;
			}
			
			cur.sector = sector;
			
			//bestimmen von prevMax:
			if(prev == null){
				cur.prevMax = cur; //nur wenn cur = trie.Min ist
			}else if(prev.count == cur.count){
				cur.prevMax = prev.prevMax;
			}else if(prev.count > cur.count){
				cur.prevMax = prev;
			}else{
				TrieLeaf max = prev.prevMax;
				while(cur.count > max.count && max != max.prevMax){
					max = max.prevMax;
				}
				if(max.count == cur.count) cur.prevMax = max.prevMax;
				else cur.prevMax = max;
			}
			prev = cur;
		}
		System.out.println("Hashtable built. Sum of " + sum);
		
		//Hier ist prev = trie.max
		prev.nextMax = prev;
		
		//Rueckrichtung:
		for(cur = prev.prev; cur != null; cur = cur.prev){
			cur.next = prev;
		
			//bestimmen von nextMax:
			if(prev.count == cur.count){
				cur.nextMax = prev.nextMax;
			}else if(prev.count > cur.count){
				cur.nextMax = prev;
			}else{
				TrieLeaf max = prev.nextMax;
				while(cur.count > max.count && max != max.nextMax){
					max = max.nextMax;
				}
				if(max.count == cur.count) cur.nextMax = max.nextMax;
				else cur.nextMax = max;
			}
			prev = cur;
		}
		
		trie.ordered = true;
		
		System.out.println("Trie ordered. Dividing up...");
		
		//die benoetigten Werte von SectorMax werden befuellt:
		sectorMax = new int[sector+1][sector+1];
		for(int i = 0; i <= sector; i++){
			for(int j = i+1; j <= sector; j++){
				if(sectorMax[i][j-1] <= sectorVal[j])
					sectorMax[i][j] = sectorVal[j];
				else
					sectorMax[i][j] = sectorMax[i][j-1];
			}
		}
		
		System.out.println("Structure built.");
	}
	
	
	private static void sectorInsert(int a[], int s){
		int i;
		for(i = 0; a[i] < s && i < a.length - 1; i++){
			if(a[i+1] < s)
				a[i] = a[i+1];
			else
				a[i] = s;
		}
		if(i == a.length - 1){
			a[i] = s;
		}
	}
	
	public QueryResult query(Trigram trigram1, Trigram trigram2){
		TrieLeaf start = leaves.find(TrigramWrapper.getString(trigram1));
		TrieLeaf end = leaves.find(TrigramWrapper.getString(trigram2));
		if(start == null)
			start = (TrieLeaf) trie.getCeilingNode(TrigramWrapper.getString(trigram1));
		if(end == null)
			end = (TrieLeaf) trie.getFloorNode(TrigramWrapper.getString(trigram2));

		if(start == null || end == null || start.sum > end.sum) return new QueryResult(0, 0);
		
		long sum = (long) end.sum - start.sum + start.count;
		
		if(start.sector != end.sector){
			return new QueryResult(sectorMax[start.sector - start.secOffset][end.sector], sum);
		}
		
		while(start.nextMax.sum <= end.prevMax.sum){
			start = start.nextMax;
			end = end.prevMax;
		}
		return new QueryResult(	(long) start.count > end.count ? start.count : end.count, sum);
	}
}
