/* Gruppe 8
 *
 * Autoren: Daniel Mock, Joshua Fuerste
 */
package grp8;

import trigrams.*;

public class Treap implements TrigramStorage {
	private TreapNode root;

	public Treap () {
		root = null;
	}

	/* 
	Funktionen fuer Interface TrigramStorage
	*/

	public void insert (TrigramAndCount trigram) {
		treapAdd (trigramToString (trigram), trigram.getCount());
	}

	public void build () {
		this.sumDown();
		//printTree(this.root, 0);
	}
	
	public QueryResult query (Trigram trigram1, Trigram trigram2) {
		
		TreapNode [] result = this.extendedSearch (trigramToString (trigram1),trigramToString (trigram2));
	 	if (result == null ) {
	 		return new QueryResult (0,0);
	 	}

	 	//berechnet summe mit sumdown
	 	long sum = result[2].sum - result[1].sum + result[1].p;
	 	QueryResult qr = new QueryResult (result[0].p, sum);
	 	return qr;
	}


	

	//Konvertiert Trigram zu String, getrennt durch Kommata
	private static String trigramToString (Trigram trigram) {
		return trigram.getW1() + "," + trigram.getW2() + "," + trigram.getW3();
	}

	//Gibt 3 Knoten zurueck:
	//0: letzter gemeinsamer knoten
	//1: fund fuer 1. trigramm (uu aufrunden)
	//2. fund fuers 2. (uu abrunden)
	private TreapNode[] extendedSearch (String s1, String s2) {
		TreapNode[] result = new TreapNode[3]; //0 = common, 1 = ergebnis 1, 2 = ergebnis2
		result[0] = this.root;
		result[1] = this.root;
		result[2] = this.root;

		//sucht den letzten gemeinsamen knoten
		//macht dabei anfang der suche fuer 1 und 2
		
		int comp = 0;
		while (result[1] == result[2]) {
			result[0] = result[1];

			comp = s1.compareTo(result[1].x);
			if (comp < 0) {
				if (result[1].left != null) {
					result[1] = result[1].left;
				} else {
					break;
				}	
			} else if (comp > 0) {
				if (result[1].right != null) {
					result[1] = result[1].right;
				} else {
					result[1] = bstSucc(result[1]);
					break;
				}
			} else {
				break;
			}

			comp = s2.compareTo(result[2].x);
			if (comp < 0) {
				if (result[2].left != null) {
					result[2] = result[2].left;
				} else {
					result[2] = bstPred (result[2]);
					break;
				}
			} else if (comp > 0) {
				if (result[2].right != null) {
					result[2] = result[2].right;
				} else {
					break;
				}
			} else {
				break;
			}
		}

		//Finde Trigram 1
		while (true) {
			comp = s1.compareTo(result[1].x);
			if (comp < 0) {
				if (result[1].left != null) {
					result[1] = result[1].left;
				} else {
					break;
				}	
			} else if (comp > 0) {
				if (result[1].right != null) {
					result[1] = result[1].right;
				} else {
					result[1] = bstSucc(result[1]);
					break;
				}
			} else {
				break;
			}
		}

		//Finde Trigram2
		while (true) {
			comp = s2.compareTo(result[2].x);
			if (comp < 0) {
				if (result[2].left != null) {
					result[2] = result[2].left;
				} else {
					result[2] = bstPred (result[2]);
					break;
				}
			} else if (comp > 0) {
				if (result[2].right != null) {
					result[2] = result[2].right;
				} else {
					break;
				}
			} else {
				break;
			}
		}

		// wenn trigram 1 und 2 nicht in DB und nichts dazwischen, null ausgeben
		if (result[1] == null || result [2] == null) {
			return null;
		}
		if ((result[1].x).compareTo(result[2].x) > 0) {
			return null;
		}

		return result;
	}

	//Fuegt in baum ein
	private  void treapAdd (String s, int count){
		TreapNode u = new TreapNode (s, count);
		this.bstIns(u);
		this.bubbleUp(u);		
		
	}

	//Stellt nach dem Einfuegen die Heapeigenschaft wieder her
	private void bubbleUp(TreapNode u) {
		while (u.parent != null && (u.parent.p < u.p )) { //|| u.parent.p == u.p && u.parent.p2 < u.p2
			if (u.parent.right == u) {
				rotateLeft(u.parent);
			} else {
				rotateRight(u.parent);
			}
		}
		if (u.parent == null) {
			this.root = u; //ACHTUNG
		}
	}

	//Speichert die Summe der Vorgaenger und sich selbst im Knoten unter sum
	private void sumDown () {
		// Random rand = new Random();
		TreapNode node = this.bstMin(this.root);
		long i = 0;
		while (node != null) {
			node.sum = i + node.p;
			i = node.sum;
			// node.p2 = rand.nextInt();
			node = this.bstSucc(node);
		}

	}

	//Fuegt wie in BST aus VL ein
	private void bstIns (TreapNode node) {
		TreapNode root = this.root, parent = null;
		while (root != null) {
			parent = root;
			if (node.x.compareTo (root.x) < 0) {
				root = root.left;
			} else {
				root = root.right;
			}
		}
		node.parent = parent;
		if (parent == null) {
			this.root = root;
		} else if (node.x.compareTo (parent.x) < 0) {
			parent.left = node;
		} else {
			parent.right = node;
		}
	}

	//Berechnet den Nachfolger im BST wie in VL
	private TreapNode bstSucc (TreapNode n) {
		TreapNode node = n;
		if (node.right != null) {
			return bstMin (node.right);
		}

		while (node.parent != null && node.parent.right == node) {
			node = node.parent;
		}
		return node.parent;
	}

	private TreapNode bstPred (TreapNode n) {
		TreapNode node = n;
		if (node.left != null) {
			return bstMax (node.left);
		}

		while (node.parent != null && node.parent.left == node) {
			node = node.parent;
		}
		return node.parent;
	}

	//Berechnet Min eines Teilbaums wie in VL
	private TreapNode bstMin (TreapNode root) {
		while (root.left != null) {
			root = root.left;
		}
		return root;
	}

	private TreapNode bstMax (TreapNode root) {
		while (root.right != null) {
			root = root.right;
		}
		return root;
	}

	public void rotateLeft(TreapNode u) {
		TreapNode w = u.right;
		w.parent = u.parent;
		if (w.parent != null) {
			if (w.parent.left == u) {
				w.parent.left = w;
			} else {
				w.parent.right = w;
			}
		}
		u.right = w.left;
		if (u.right != null) {
			u.right.parent = u;
		}
		u.parent = w;
		w.left = u;
		if (u == this.root) { //ACHTUNG r
			this.root = w; 
			this.root.parent = null; 
		}
	}

	private void rotateRight(TreapNode u) {
		TreapNode w = u.left;
		w.parent = u.parent;
		if (w.parent != null) {
			if (w.parent.left == u) {
				w.parent.left = w;
			} else {
				w.parent.right = w;
			}
		}
		u.left = w.right;
		if (u.left != null) {
			u.left.parent = u;
		}
		u.parent = w;
		w.right = u;
		if (u == this.root) {  //ACHTUNG r
			this.root = w; 
			this.root.parent = null; 
		}
	}

	private class TreapNode {
		String x; // string des Trigrams
		int p;	// prio des Treaps, count des trigrams
		int p2; // zweite Prio, falls erste gleich ist.
		long sum; 	// Summe des Teilbaums auch sich selbst!
		TreapNode parent;
		TreapNode left;
		TreapNode right;

		TreapNode (String key, int count) {
			this.x = key;
			this.p = count;
			this.p2 = 0;
			this.parent = null;
			this.left = null;
			this.right = null;
		}

		
	}

}


