/* Gruppe 9
 *
 * Autoren: anonym
 */
package grp9;

import java.util.ArrayList;

import trigrams.*;

public class MyTrigramStorage implements TrigramStorage
{

	private ArrayList<MyTrigram> input;
	private MyTrigram[] trigrams;

	public MyTrigramStorage()
	{
		input = new ArrayList<MyTrigram>();
	}
	
	@Override
	public void insert(TrigramAndCount trigram)
	{
		input.add(new MyTrigram(trigram));
	}

	@Override
	public void build()
	{
		trigrams = new MyTrigram[0];
		trigrams = input.toArray(trigrams);
		input = null;  //reduce memory-use
		
		//sort
		quicksort(0,trigrams.length-1);
		
		//calc cumulated counts
		int count = 0;
		for(int i = 0; i < trigrams.length; i++)
		{
			count += trigrams[i].getCount();
			trigrams[i].setCumulatedCount(count);
		}
		
		//calc next Maxs
		for(int i = 0; i < trigrams.length; i++)
		{
			int j = i+1;
			while(j<trigrams.length && trigrams[i].getCount() >= trigrams[j].getCount())
			{
				j++;
			}
			if(j == trigrams.length)
			{
				trigrams[i].setNextMax(-1);
			}
			else
			{
				trigrams[i].setNextMax(j);
			}
		}
	}
	
	private void quicksort(int left, int right)
	{
		if(left < right)
		{
			int i = partition(left, right);
			
			quicksort(left, i-1);
			quicksort(i+1, right);
		}
	}
	
	private int partition(int left, int right)
	{
		int ppos = right;
		Trigram pivot = trigrams[ppos];
		while(true)
		{
			while(left < right && trigrams[left].compareTo(pivot) < 0)
				left++;
			while (left < right && trigrams[right].compareTo(pivot) >= 0)
				right--;
			
			if(left >= right)
				break;
			
			swap(left, right);
		}
			
		swap(left, ppos);
			
		return left;
	}
	
	private void swap(int i, int j)
	{
		MyTrigram tmp = trigrams[i];
		trigrams[i] = trigrams[j];
		trigrams[j] = tmp;
	}

	@Override
	public QueryResult query(Trigram trigram1, Trigram trigram2)
	{
		int sum;
		int max;

		//get left pos
		int pos = 0;
		int l = 0;
		int r = trigrams.length-1;
		while(l<=r)
		{
			pos = (l+r)/2;
			int cmp = trigrams[pos].compareTo(trigram1);
			if(cmp == 0)
				break;
			if(cmp > 0)
				r = pos-1;
			else
				l = pos+1;
		}
		int left = pos;
		while(left <= trigrams.length-1 && trigrams[left].compareTo(trigram1) < 0)
			left++;
		
		//get right pos
		l = 0;
		r = trigrams.length-1;
		while(l<=r)
		{
			pos = (l+r)/2;
			int cmp = trigrams[pos].compareTo(trigram2);
			if(cmp == 0)
				break;
			if(cmp > 0)
				r = pos-1;
			else
				l = pos+1;
		}
		int right = pos;
		while(right >= 0 && trigrams[right].compareTo(trigram2) > 0)
			right--;

		if(right < left)
			return new QueryResult(0,0);

		//calc sum
		if(left > 0)
			sum = trigrams[right].getCumulatedCount() - trigrams[left-1].getCumulatedCount();
		else
			sum = trigrams[right].getCumulatedCount();
		
		//calc max
		MyTrigram maxTrigram = trigrams[left];
		while(maxTrigram.getNextMax() != -1 && maxTrigram.getNextMax() <= right)
		{
			maxTrigram = trigrams[maxTrigram.getNextMax()];
		}
		max = maxTrigram.getCount();
		
		return new QueryResult(max, sum);
	}
}
